open Xlib
open Xstring
open Interfaces

class virtual abstract_terms_definition _gmod : terms_definition =
object (self)
	(** the grammar module currently processed *)
	val gmod : gmodule = _gmod
	
	(** the target grammar actually processed *)
	val mutable target_grammar : grammar option = None
	method private set_target (g : grammar) : unit = target_grammar <- Some g
	method get_target : grammar = Xlib.some_of target_grammar

	(** definition title *)
	method virtual title : string

	(** extra definition lines to be inserted before the grammar definitions *)
	method coq_predefs : string list = []

	(** extra definition lines to be inserted after the constructor cases *)
	method aux_defs : string list = []

	(** the category of generated index case *)
	method virtual index_type : string
	
	(** category names decoration (to avoid conflict) *)
	method decorate_type_name (name : string) : string = name
	
	(** constructors names decoration (to avoid conflict) *)
	method decorate_constr_name (name : string) : string = name
	
	(** generation of shifts *)
	method virtual shift_terms_gen : (grammar * string) -> string
	
	(** generation of parameters *)
	method parameter_terms_gen (p : parameter) : string =
		let shifts =
			String.concat " "
				(List.map self#shift_terms_gen p#get_shifts) in
		let shifts_final = if shifts = "" then "" else shifts ^ " " in
		shifts_final ^ "(" ^ p#get_name ^ " : " ^ self#decorate_type_name p#get_type#get_name ^ ")"
	
	(** generation of constructors *)
	method constructor_terms_gen (c : constructor) : string =
		let subst =
			("#constr#", self#decorate_constr_name c#get_name) ::
			("#ind#", c#get_index_name) ::
			("#index_type#", self#index_type) ::
			("#params#",
				String.concat
					" " (List.map self#parameter_terms_gen c#get_params)) ::
			[]
		and text =
			if c#is_index
			then
				"  | #constr# (#ind# : #index_type#)"
			else
				"  | #constr# #params#"
		in
		Xstring.replace subst text
	
	(** generation of grammars *)
	method grammar_terms_gen (prefix : string) (g : grammar) : string list =
		let _ = self#set_target g in
		let subst =
			("#Inductive#", prefix) ::
			("#raw_type#", g#get_name) ::
			("#type#", self#decorate_type_name g#get_name) ::
			[]
		and substs =
			("#constr_lines#",
				List.map self#constructor_terms_gen g#get_constrs) ::
			("#aux_lines#", self#aux_defs) ::
			[]
		and text =
			"  #Inductive# #type# : Type :=" ::
			"#constr_lines#" ::
			"#aux_lines#" ::
			[] in
		Xstring.replaces subst substs text
end

class virtual abstract_fun_definition _gmod : fun_definition =
object (self)
	(** the grammar module currently processed *)
	val gmod : gmodule = _gmod
	
	(** the source grammar actually processed *)
	val mutable source_grammar : grammar option = None
	method private set_source (g : grammar) : unit = source_grammar <- Some g
	method get_source : grammar = Xlib.some_of source_grammar
	
	(** the target grammar actually processed *)
	val mutable target_grammar : grammar option = None
	method private set_target (g : grammar) : unit = target_grammar <- Some g
	method get_target : grammar = Xlib.some_of target_grammar
	
	(** the definition title *)
	method virtual title : string
	
	(** extra definition lines to be inserted before the grammar definitions *)
	method coq_predefs : string list = []

	(** extra definition lines to be inserted after the constructor cases *)
	method aux_defs : string list = []

	(** function parameters list (name * type_name) *)
	method virtual fun_params : (string * string) list	
	
	(** return type of the function *)
	method virtual return_type : string
	
	(** name of the function in a given target grammar *)
	method virtual fun_name_type_name : string -> string
	
	(** checks the validity of the function *)
	method is_valid : bool =
		(List.mem (Some self#get_source) self#get_source_grams) &&
		(List.mem self#get_target (self#get_target_grams gmod#get_all_grams))
	
	(** name of the function *)
	method fun_name : string =
		self#fun_name_type_name self#get_target#get_name
	
	(** category names decoration (to avoid conflict) *)
	method decorate_type_name : string -> string = fun x -> x
	
	(** constructors names decoration (to avoid conflict) *)
	method decorate_constr_name : string -> string = fun x -> x
	
	(** the source grammars for which generetion is needed. *)
	method virtual get_source_grams : grammar option list
	
	(** the target (processed) grammars for which generetion is needed.
	* By default, every grammar with the source grammar in its throughs. *)
	method virtual get_target_grams : grammar list -> grammar list
	
	(** check if the type of the parameter is a self recursion *)
	method is_self_rec (p : parameter) : bool =
		p#get_type = self#get_target
	
	(** check if the type of the parameter is a back recursion *)
	method is_back_rec (p : parameter) : bool =
		p#get_type = self#get_source
	
	(** check if the type of the parameter is a throughs recursion *)
	method is_throughs_rec (p : parameter) : bool =
		List.mem self#get_source (p#get_type#get_throughs)
	
	(** check if the type of the parameter is a module recursion *)
	method is_module_rec (p : parameter) : bool =
		gmod#is_gmodule_gram p#get_type#get_name
	
	(** generation of shifts *)
	method virtual shift_fun_gen : (grammar * string) -> string
	
	(** generation of parameters *)
	method virtual parameter_function_gen : parameter -> string
	
	(** generation of constructors rhs : index case *)
	method virtual index_constructor_rhs_function_gen : constructor -> string list
	
	(** generation of constructors rhs : non index case *)
	method constructor_rhs_function_gen (c : constructor) : string list =
		[ String.concat " "
			(c#get_name :: (List.map (fun p -> self#parameter_function_gen p) c#get_params)) ]
	
	(** generation of constructors rhs *)
	method rhs_function_gen (c : constructor) : string list =
		if c#is_index
		then self#index_constructor_rhs_function_gen c
		else self#constructor_rhs_function_gen c
	
	(** generation of constructors *)
	method constructor_function_gen (c : constructor) : string list =
		let param_names =
			String.concat
				" " (List.map (fun p -> p#get_name) c#get_params) in
		let subst =
			("#constr#", self#decorate_constr_name c#get_name) ::
			("#params#", param_names) ::
			[]
		and substs =
			("#rhs#", List.map (fun s -> "        " ^ s) (self#rhs_function_gen c)) ::
			[]
		and text =
			"      | #constr# #params# =>" ::
			"#rhs#" ::
			[] in
		Xstring.replaces subst substs text
	
	(** generation of constructors fun equation *)
	method constructor_function_equation_gen (c : constructor) : string list =
		let typed_constr_params =
			String.concat
				" " (List.map (fun p -> "(" ^ p#get_name ^ " : " ^ p#get_type#get_name ^ ")") c#get_params) in
		let constr_param_names =
			String.concat
				" " (List.map (fun p -> p#get_name) c#get_params) in
		let subst =
			("#constr#", self#decorate_constr_name c#get_name) ::
			("#typed_fun_params#",
				String.concat " "
					(List.map (fun (x, t) -> "(" ^ x ^ " : " ^ t ^ ")") self#fun_params)) ::
			("#fun_params#",
				String.concat " " (List.map fst self#fun_params)) ::
			("#typed_constr_params#", typed_constr_params) ::
			("#constr_params#", constr_param_names) ::
			("#fun_name#", self#fun_name) ::
			("#lemma_name#", self#fun_name ^ "_" ^ c#get_name) ::
			[]
		and substs =
			("#rhs#", List.map (fun s -> "        " ^ s) (self#rhs_function_gen c)) ::
			[]
		and text =
			"  Lemma #lemma_name# :" ::
			"    forall #typed_fun_params# #typed_constr_params#," ::
			"      #fun_name# #fun_params# (#constr# #constr_params#) =" ::
			"#rhs#" ::
			"  ." ::
			"  Proof." ::
			"    intros; simpl; auto." ::
			"  Qed." ::
			"" ::
			"  Hint Rewrite #lemma_name# : #database#." ::
			[] in
		Xstring.replaces subst substs text
	
	(** generation of function for a given target grammar *)
	method grammar_function_gen (prefix : string) (target : grammar) : string list =
		let _ = self#set_target target in
		let subst =
			("#Fixpoint#", prefix) ::
			("#raw_target_type#", target#get_name) ::
			("#target_type#", self#decorate_type_name target#get_name) ::
			("#source_type#",
				match source_grammar with
					Some g -> self#decorate_type_name g#get_name
				| None -> "") ::
			("#fun_name#", self#fun_name) ::
			("#return_type#", self#return_type) ::
			("#fun_params#",
				String.concat " "
					(List.map (fun (x, t) -> "(" ^ x ^ " : " ^ t ^ ")") self#fun_params)) ::
			[]
		and substs =
			("#constr_lines#",
				List.concat (List.map self#constructor_function_gen target#get_constrs)) ::
			("#aux_lines#", self#aux_defs) ::
			[]
		and text =
			"  #Fixpoint# #fun_name# #fun_params# (_arg : #target_type#) { struct _arg } : #return_type# :=" ::
			"    match _arg with" ::
			"#constr_lines#" ::
			"#aux_lines#" ::
			"    end" ::
			[]
		in
		Xstring.replaces subst substs text
end
