open Debug
open Xlib
open Xlist
open Xcoq
open Interfaces

class virtual_property _gmod : property =
  object (self)
      (** the grammar module currently processed *)
    val gmod : gmodule = _gmod
	
	(** the node actually processed *)
    val mutable target_node : node option = None
	(** the source grammars actually processed *)
    val mutable source_grammar : grammar option = None
    val mutable alt_source_grammar : grammar option = None
	(** the target grammar actually processed *)
    val mutable target_grammar : grammar option = None
	
    method set_node (n : node) : unit = target_node <- Some n
    method set_source (g : grammar) : unit = source_grammar <- Some g
    method set_alt_source (g : grammar) : unit = alt_source_grammar <- Some g
    method set_target (g : grammar) : unit = target_grammar <- Some g
    method get_node : node = Xlib.some_of target_node
    method has_source : bool = source_grammar <> None
    method get_source : grammar = Xlib.some_of source_grammar
    method get_alt_source : grammar = Xlib.some_of alt_source_grammar
    method get_target : grammar = Xlib.some_of target_grammar
	
	(** the source grammars for which generation is needed. *)
    method get_source_grams : (grammar * grammar option) list =
      failwith "definitions_virtual.virtual_fun_definition : no get_source_grams specified"
	
	(** the target grammars to process among a given grammar list *)
    method get_target_grams : grammar list -> grammar list =
      failwith "definitions_virtual.virtual_fun_definition : no get_source_grams specified"
	
	(** Check if an induction for a source and target grammars is valid. *)
    method is_valid_induction (source : grammar) (target : grammar) : bool =
      (List.mem source (List.map (fun (g, _) -> g) self#get_source_grams)) &&
      (List.mem target (self#get_target_grams gmod#get_all_grams))
	
	(** name of the lemma for a given source name *)
    method lemma_name_type_name : string -> string =
      failwith "properties_abstract.virtual_property : no lemma_name specified."
	
	(** name of the lemma *)
    method lemma_name : string =
      self#lemma_name_type_name self#get_target#get_name
	
	(** the lemma statement *)
    method lemma_header : string list =
      failwith "properties_abstract.virtual_property : no header specified"
	
	(** the lemma proof script *)
    method lemma_proof : string list =
      failwith "properties_abstract.virtual_property : no proof script specified"
	
	(** extra definition lines to be inserted after the property *)
    method aux_defs : string list = []

	(** the tactic template : pattern, body and body postpone *)
    method tactic_template : string * (string list * string list) =
      ("", ([], []))
	
	(** index tactic to use when performing a structural induction *)
    method get_index_tac : string =
      failwith "properties_abstract.virtual_property : no get_index_tac specified"
	
	(** extra substitution to perform to definition inside the property *)
    method subst : (string * string) list = []
	
	(** generation for a given module *)
    method generate : string list * (string * (string list * string list)) list =
      let one_source_gen (source, alt_source) =
	let _ = source_grammar <- Some source
	and _ = alt_source_grammar <- alt_source in
	gmod#property_gen (self :> property) in
      let (lemmas, tactics) =
	List.split
	  (List.map one_source_gen self#get_source_grams) in
      (Xlist.concat_with [""] lemmas, List.concat tactics)
  end

class virtual_property_pack _gmod : property_pack =
object (self)
	(** the grammar module currently processed *)
	val gmod : gmodule = _gmod

	(** the title to insert before the definitions *)
	method get_title : string =
		failwith "properties_abstract.virtual_property_pack : no title specified"
	
	(** the list of properties of this pack *)
	method properties : property list = []
	
	(** the tactic name of the tactic associated with this pack,
	* or None if no tactic is required *)
	method tactic_name : string option = None
	
	(** generation for a given module *)
	method generate : string list =
		let (lemmas, tactics) =
			List.split
				(List.map (fun p -> p#generate) self#properties) in
		let all_lemmas = Xlist.concat_with [""] lemmas in
		let all_tactics_aggr =
			Xcoq.tactics_aggregate (List.concat tactics) in
		let all_tactics_merge = Xcoq.tactics_merge all_tactics_aggr in
		let all_tactics =
			match self#tactic_name with
					None -> []
				| Some tac_name ->
						"" ::
						("  Ltac " ^ tac_name ^ " :=") ::
						"    repeat" ::
						"      match goal with" ::
						all_tactics_merge @
						"      end; #crush#." ::
						[] in
		Xcoq.coq_title 3 self#get_title @ [""] @
		all_lemmas @ all_tactics
end

class virtual_property_chapter _gmod : property_chapter =
object (self)
	(** the grammar module currently processed *)
	val gmod : gmodule = _gmod

	(** the title to insert before the definitions *)
	method get_title : string =
		failwith "properties_abstract.virtual_property_chapter : no title specified"
	
	(** the list of property packs of this chapter *)
	method property_packs : property_pack list = []
	method packs_tactic_names : string list =
		let rec aux pl =
			match pl with
			  [] -> []
			| p :: pl ->
				(match p#tactic_name with
				    None -> aux pl
					| Some s -> s :: aux pl) in
		aux self#property_packs
	
	(** auxiliary coq lines to add before and after this chapter *)
	method before_defs : string list = []
	method after_defs : string list = []
	
	(** generation for a given module *)
	method generate : string list =
		Xcoq.coq_title 2 self#get_title @ [""] @
		self#before_defs @ [""] @
		Xlist.concat_with [""]
			(List.map (fun p -> p#generate) self#property_packs) @ [""] @
		self#after_defs
end
